from numpy import add


class KronSumMean(object):
    def __init__(self, A, B):
        self._A = A
        self._B = B

    def value_reduce(self, values):
        return None

    def derivative_reduce(self, derivatives):
        return None



from numpy import nan
from numpy import eye
from numpy import dot
from numpy import copyto
from numpy.random import RandomState

from limix_inference.mean import KronSumMean

def _gower_normalization(K, out=None):
    """Perform Gower normalizion on covariance matrix K.

    The rescaled covariance matrix has sample variance of 1.
    """
    c = (K.shape[0] - 1) / (K.trace() - K.mean(0).sum())
    if out is None:
        return c * K

    copyto(out, K)
    out *= c

def test_kron_two():
    random = RandomState(1)

    # define phenotype
    N = 10
    P = 3
    Y = random.randn(N,P)

    # pheno with missing data
    Ym = Y.copy()
    Im = random.rand(N, P) < 0.2
    Ym[Im] = nan

    # # define fixed effects
    F = 1.*(random.rand(N,2) < 0.5)
    A = eye(P)

    # mean = MeanKronSum(Y, F=F, A=A)
    mean = KronSumMean(F, A)

    # mean_m = MeanKronSum(Ym, F=F, A=A)

    # define row covariance
    f = 10
    X = 1.*(random.rand(N, f)<0.2)
    R = _gower_normalization(dot(X,X.T))
    R+= 1e-4 * eye(N)

    # define col covariances
    # Cg = FreeFormCov(P)
    # Cn = FreeFormCov(P)
    # Cg.setRandomParams()
    # Cn.setRandomParams()
    #
    # # define covariance matrices
    # covar1 = KronCov(Cg, R)
    # covar2 = KronCov(Cn, eye(N))
    # covar  = SumCov(covar1,covar2)
    #
    # # define covariance matrice with missing data
    # Iok = (~Im).reshape(N * P, order='F')
    # covar1_m = KronCov(copy.copy(Cg), R, Iok=Iok)
    # covar2_m = KronCov(copy.copy(Cn), eye(N), Iok=Iok)
    # covar_m  = SumCov(covar1_m,covar2_m)
    #
    # # define gp
    # self._gp = GP(covar=covar, mean=mean)
    # self._gpm = GP(covar=covar_m, mean=mean_m)
    # self._gp2ks = GP2KronSum(Y=Y, F=F, A=A, Cg=Cg, Cn=Cn, R=R)



from __future__ import division

from numpy import exp, log, pi

from numpy_sugar.special import normal_cdf, normal_icdf


class Link(object):
    def __init__(self):
        super(Link, self).__init__()

    def value(self, x):
        raise NotImplementedError

    def inv(self, x):
        raise NotImplementedError

    @property
    def latent_variance(self):
        raise NotImplementedError


class LogitLink(Link):
    def __init__(self):
        super(LogitLink, self).__init__()

    def value(self, x):
        return log(x / (1 - x))

    def inv(self, x):
        return 1 / (1 + exp(-x))

    @property
    def latent_variance(self):
        return pi**2 / 3.0


class ProbitLink(Link):
    def __init__(self):
        super(ProbitLink, self).__init__()

    def value(self, x):
        return normal_icdf(x)

    def inv(self, x):
        return normal_cdf(x)

    @property
    def latent_variance(self):
        return 1.0


class LogLink(Link):
    def __init__(self):
        super(LogLink, self).__init__()

    def value(self, x):
        return log(x)

    def inv(self, x):
        return exp(x)



"""
*******************************
Generalized Linear Mixed Models
*******************************

Example
^^^^^^^

.. doctest::

    >>> from limix_inference.random import bernoulli_sample
    >>> from limix_inference.glmm import ExpFamEP
    >>> from limix_inference.lik import BernoulliProdLik
    >>> from limix_inference.link import LogLink
    >>> from numpy_sugar.linalg import economic_qs_linear
    >>> from numpy.random import RandomState
    >>>
    >>> offset = 0.2
    >>> random = RandomState(0)
    >>> G = random.randn(100, 200)
    >>> (Q0, Q1), S0 = economic_qs_linear(G)
    >>> y = bernoulli_sample(offset, G, random_state=random)
    >>> covariates = random.randn(100, 1)
    >>> lik = BernoulliProdLik(LogLink)
    >>> lik.outcome = y
    >>> glmm = ExpFamEP(lik, covariates, Q0, Q1, S0)
    >>> glmm.learn(progress=False)
    >>> '%.4f' % glmm.lml()
    '-69.0610'

Expectation propagation
^^^^^^^^^^^^^^^^^^^^^^^

.. autoclass:: ExpFamEP
    :members: covariates_variance, genetic_variance, environmental_variance,
              heritability, K, m, beta, v, delta, lml, learn, fixed_ep
"""

from ._ep import ExpFamEP



from __future__ import division

from numpy import sqrt, std, ascontiguousarray, arange

from ..link import LogitLink
from ..link import LogLink

from ..lik import BernoulliProdLik
from ..lik import BinomialProdLik
from ..lik import PoissonProdLik

from ..mean import OffsetMean
from ..mean import LinearMean
from ..mean import SumMean

from ..cov import LinearCov
from ..cov import SumCov
from ..cov import EyeCov

from ._glmm import GLMMSampler


def bernoulli_sample(offset,
                     G,
                     heritability=0.5,
                     causal_variants=None,
                     causal_variance=0,
                     random_state=None):
    r"""Bernoulli likelihood sampling.

    Sample according to

    .. math::

        \mathbf y \sim \prod_{i=1}^n
        \text{Bernoulli}(\mu_i = \text{logit}(z_i))
        \mathcal N(~ o \mathbf 1 + \mathbf a^\intercal \boldsymbol\alpha;
        ~ (h^2 - v_c)\mathrm G^\intercal\mathrm G +
        (1-h^2-v_c)\mathrm I ~)

    using the canonical Logit link function to define the conditional Bernoulli
    mean :math:`\mu_i`.

    The causal :math:`\mathbf a` covariates and the corresponding effect-sizes
    are randomly draw according to the following idea. The ``causal_variants``,
    if given, are first mean-zero and std-one normalized and then having
    its elements divided by the squared-root the the number of variances::

        causal_variants = _stdnorm(causal_variants, axis=0)
        causal_variants /= sqrt(causal_variants.shape[1])

    The causal effect-sizes :math:`\boldsymbol\alpha` are draw from
    :math:`\{-1, +1\}` and subsequently normalized for mean-zero and std-one""


    """
    link = LogitLink()
    mean, cov = _mean_cov(offset, G, heritability, causal_variants,
                          causal_variance, random_state)
    lik = BernoulliProdLik(link)
    sampler = GLMMSampler(lik, mean, cov)

    return sampler.sample(random_state)


def binomial_sample(ntrials,
                    offset,
                    G,
                    heritability=0.5,
                    causal_variants=None,
                    causal_variance=0,
                    random_state=None):
    """Binomial likelihood sampling."""
    link = LogitLink()
    mean, cov = _mean_cov(offset, G, heritability, causal_variants,
                          causal_variance, random_state)
    lik = BinomialProdLik(ntrials, link)
    sampler = GLMMSampler(lik, mean, cov)

    return sampler.sample(random_state)


def poisson_sample(offset,
                   G,
                   heritability=0.5,
                   causal_variants=None,
                   causal_variance=0,
                   random_state=None):
    """Poisson likelihood sampling."""
    mean, cov = _mean_cov(offset, G, heritability, causal_variants,
                          causal_variance, random_state)
    link = LogLink()
    lik = PoissonProdLik(link)
    sampler = GLMMSampler(lik, mean, cov)

    return sampler.sample(random_state)


def _causal_mean(causal_variants, causal_variance, random):
    causal_variants = _stdnorm(causal_variants, axis=0)
    causal_variants /= sqrt(causal_variants.shape[1])
    p = causal_variants.shape[1]
    directions = random.randn(p)
    directions[directions < 0.5] = -1
    directions[directions >= 0.5] = +1
    s = std(directions)
    if s > 0:
        directions /= s
    directions *= sqrt(causal_variance)
    directions -= directions.mean()
    mean = LinearMean(p)
    mean.set_data((causal_variants, ), 'sample')
    mean.effsizes = directions
    return mean


def _mean_cov(offset, G, heritability, causal_variants, causal_variance,
              random_state):
    G = ascontiguousarray(G, dtype=float)
    nsamples = G.shape[0]
    G = _stdnorm(G, axis=0)

    G /= sqrt(G.shape[1])

    mean1 = OffsetMean()
    mean1.offset = offset

    cov1 = LinearCov()
    cov2 = EyeCov()
    cov = SumCov([cov1, cov2])

    mean1.set_data(nsamples, 'sample')
    cov1.set_data((G, G), 'sample')
    a = arange(nsamples)
    cov2.set_data((a, a), 'sample')

    cov1.scale = heritability - causal_variance
    cov2.scale = 1 - heritability - causal_variance

    means = [mean1]
    if causal_variants is not None:
        means += [_causal_mean(causal_variants, causal_variance, random_state)]

    mean = SumMean(means)

    return mean, cov


def _stdnorm(X, axis=None, out=None):
    X = ascontiguousarray(X, dtype=float)
    if out is None:
        out = X.copy()

    m = out.mean(axis)
    s = out.std(axis)
    ok = s > 0

    out -= m

    if out.ndim == 1:
        if s > 0:
            out /= s
    else:
        out[..., ok] /= s[ok]

    return out



from numpy.random import RandomState
from numpy.testing import assert_almost_equal

from optimix import check_grad

from limix_inference.mean import LinearMean


def test_value():
    random = RandomState(0)
    mean = LinearMean(5)
    effsizes = random.randn(5)
    mean.effsizes = effsizes

    x = random.randn(5)
    assert_almost_equal(mean.value(x), -0.956409566703)


def test_gradient():
    random = RandomState(1)
    mean = LinearMean(5)
    effsizes = random.randn(5)
    mean.effsizes = effsizes

    x = random.randn(5)

    def func(x0):
        mean.effsizes[0] = x0[0]
        return mean.value(x)

    def grad(x0):
        mean.effsizes[0] = x0[0]
        return [mean.derivative_effsizes(x)[0]]

    assert_almost_equal(check_grad(func, grad, [1.2]), 0, decimal=6)


if __name__ == '__main__':
    __import__('pytest').main([__file__, '-s'])



from __future__ import division

import logging


def make_sure_reasonable_conditioning(S):
    max_cond = 1e6
    cond = S.max() / S.min()
    if cond > max_cond:
        eps = (max_cond * S.min() - S.max()) / (1 - max_cond)
        logger = logging.getLogger(__name__)
        logger.warning("The covariance matrix's conditioning number" +
                       " is too high: %e. Summing %e to its eigenvalues and " +
                       "renormalizing for a better conditioning number.", cond,
                       eps)
        m = S.mean()
        S += eps
        S *= m / S.mean()



from __future__ import division

from numpy import exp
from numpy.random import RandomState
from numpy.testing import assert_almost_equal

from optimix import check_grad

from limix_inference.lmm import SlowLMM
from limix_inference.cov import LinearCov
from limix_inference.mean import OffsetMean
from limix_inference.mean import LinearMean


def test_slowlmm_value_1():
    random = RandomState(94584)
    N = 50
    X = random.randn(N, 100)
    offset = 0.5

    mean = OffsetMean()
    mean.offset = offset
    mean.fix('offset')
    mean.set_data(N)

    cov = LinearCov()
    cov.scale = 1.0
    cov.set_data((X, X))

    y = random.randn(N)

    lmm = SlowLMM(y, mean, cov)
    assert_almost_equal(lmm.feed().value(), -153.623791551399108)


def test_slowlmm_value_2():
    random = RandomState(94584)
    N = 50
    X1 = random.randn(N, 3)
    X2 = random.randn(N, 100)

    mean = LinearMean(3)
    mean.set_data(X1)

    cov = LinearCov()
    cov.scale = 1.0
    cov.set_data((X2, X2))

    y = random.randn(N)

    lmm = SlowLMM(y, mean, cov)
    assert_almost_equal(lmm.feed().value(), -153.091074766)

    mean.effsizes = [3.4, 1.11, -6.1]
    assert_almost_equal(lmm.feed().value(), -178.273116338)


def test_regression_gradient():
    random = RandomState(94584)
    N = 50
    X = random.randn(N, 100)
    offset = 0.5

    mean = OffsetMean()
    mean.offset = offset
    mean.fix('offset')
    mean.set_data(N)

    cov = LinearCov()
    cov.scale = 1.0
    cov.set_data((X, X))

    y = random.randn(N)

    lmm = SlowLMM(y, mean, cov)

    def func(x):
        cov.scale = exp(x[0])
        return lmm.value(mean.feed().value(), cov.feed().value())

    def grad(x):
        cov.scale = exp(x[0])
        return lmm.gradient(mean.feed().value(),
                            cov.feed().value(),
                            mean.feed().gradient(), cov.feed().gradient())

    assert_almost_equal(check_grad(func, grad, [0]), 0)


def test_maximize_1():
    random = RandomState(94584)
    N = 50
    X = random.randn(N, 100)
    offset = 0.5

    mean = OffsetMean()
    mean.offset = offset
    mean.fix('offset')
    mean.set_data(N)

    cov = LinearCov()
    cov.scale = 1.0
    cov.set_data((X, X))

    y = random.randn(N)

    lmm = SlowLMM(y, mean, cov)
    m = mean.feed().value()
    K = cov.feed().value()
    assert_almost_equal(lmm.value(m, K), -153.62379155139911)

    lmm.feed().maximize()
    assert_almost_equal(lmm.feed().value(), -79.899212241487518)


def test_maximize_2():
    random = RandomState(94584)
    N = 50
    X = random.randn(N, 100)
    offset = 0.5

    mean = OffsetMean()
    mean.offset = offset
    mean.set_data(N)

    cov = LinearCov()
    cov.scale = 1.0
    cov.set_data((X, X))

    y = random.randn(N)

    lmm = SlowLMM(y, mean, cov)
    lmm.feed().maximize()
    assert_almost_equal(lmm.feed().value(), -79.365136339619610)


if __name__ == '__main__':
    __import__('pytest').main([__file__, '-s'])



from __future__ import division

from numpy import append, asarray, concatenate, dot, empty, log, newaxis, sum

from numpy_sugar.linalg import solve


class NormalLikTrick(object):
    r"""Necessary data for fast lml computation over many fixed-effects."""

    def __init__(self, M, Q0, Q1, yTQ0, yTQ1, diag0, diag1, a0, a1):
        self.M = M
        self.Q0 = Q0
        self.Q1 = Q1
        self.yTQ0 = yTQ0
        self.yTQ1 = yTQ1
        self.diag0 = diag0
        self.diag1 = diag1
        self.a0 = a0
        self.a1 = a1

    def fast_scan(self, markers):
        r"""LMLs of markers by fitting scale and fixed-effect sizes parameters.

        The likelihood is given by

        .. math::

            \mathcal N\big(~\mathbf y ~|~ \boldsymbol\beta^{\intercal}
            [\mathrm M ~~ \tilde{\mathrm M}],  s \mathrm K~\big),

        where :math:`s` is the scale parameter and :math:`\boldsymbol\beta` is the
        fixed-effect sizes; :math:`\tilde{\mathrm M}` is a marker to be scanned.
        """

        assert markers.ndim == 2

        nc = self.M.shape[1]
        M = self.M

        MTQ0 = dot(M.T, self.Q0)
        mTQ0 = dot(markers.T, self.Q0)

        MTQ1 = dot(M.T, self.Q1)
        mTQ1 = dot(markers.T, self.Q1)

        yTQ0diag0 = self.yTQ0 / self.diag0
        yTQ1diag1 = self.yTQ1 / self.diag1

        b0 = yTQ0diag0.dot(MTQ0.T)
        b0m = yTQ0diag0.dot(mTQ0.T)

        MTQ0diag0 = MTQ0 / self.diag0
        MTQ1diag1 = MTQ1 / self.diag1

        c0_00 = MTQ0diag0.dot(MTQ0.T)
        c0_01 = MTQ0diag0.dot(mTQ0.T)
        c0_11 = sum((mTQ0 / self.diag0) * mTQ0, axis=1)

        b1 = yTQ1diag1.dot(MTQ1.T)
        b1m = yTQ1diag1.dot(mTQ1.T)

        c1_00 = MTQ1diag1.dot(MTQ1.T)
        c1_01 = MTQ1diag1.dot(mTQ1.T)
        c1_11 = sum((mTQ1 / self.diag1) * mTQ1, axis=1)

        C0 = empty((nc + 1, nc + 1))
        C0[:-1, :-1] = c0_00

        C1 = empty((nc + 1, nc + 1))
        C1[:-1, :-1] = c1_00

        n = markers.shape[0]
        lmls = empty(markers.shape[1])
        LOG2PI = 1.837877066409345339081937709124758839607238769531250
        lmls[:] = -n * LOG2PI - n
        lmls[:] += -sum(log(self.diag0)) - (n - len(self.diag0)
                                            ) * log(self.diag1)

        for i in range(markers.shape[1]):

            b11m = append(b1, b1m[i])
            b00m = append(b0, b0m[i])

            nominator = b11m - b00m

            C0[:-1, -1] = c0_01[:, i]
            C1[:-1, -1] = c1_01[:, i]

            C0[-1, :-1] = C0[:-1, -1]
            C1[-1, :-1] = C1[:-1, -1]

            C0[-1, -1] = c0_11[i]
            C1[-1, -1] = c1_11[i]

            denominator = C1 - C0

            beta = solve(denominator, nominator)

            p0 = self.a1 - 2 * b11m.dot(beta) + beta.dot(C1.dot(beta))
            p1 = self.a0 - 2 * b00m.dot(beta) + beta.dot(C0).dot(beta)

            scale = (p0 + p1) / n

            lmls[i] -= n * log(scale)

        lmls /= 2
        return lmls



from __future__ import division

from numpy.random import RandomState
from numpy_sugar.linalg import sum2diag
from numpy_sugar import epsilon
from numpy_sugar.random import multivariate_normal


class GLMMSampler(object):
    def __init__(self, lik, mean, cov):
        self._lik = lik
        self._mean = mean
        self._cov = cov

    def sample(self, random_state=None):
        if random_state is None:
            random_state = RandomState()

        m = self._mean.feed('sample').value()
        K = self._cov.feed('sample').value()

        sum2diag(K, +epsilon.small, out=K)
        u = multivariate_normal(m, K, random_state)
        sum2diag(K, -epsilon.small, out=K)

        return self._lik.sample(u, random_state)



from numpy import zeros
from numpy import ascontiguousarray

from optimix import Function
from optimix import Vector


class LinearMean(Function):
    r"""Linear mean function.

    The mathematical representation is

    .. math::

        f(\mathbf x) = \mathbf x^\intercal \boldsymbol\alpha

    where :math:`\boldsymbol\alpha` is a vector of effect sizes.
    """
    def __init__(self, size):
        Function.__init__(self, effsizes=Vector(zeros(size)))

    def value(self, x):
        r"""Mean function evaluated at `x`.

        Args:
            x (array_like): sample or samples.

        Returns:
            :math:`\mathbf x^\intercal \boldsymbol\alpha`.
        """
        return x.dot(self.get('effsizes'))

    def derivative_effsizes(self, x):
        r"""Linear mean function derivative.

        Args:
            size (int): sample size.

        Returns:
            :math:`\mathbf x`.
        """
        return x

    @property
    def effsizes(self):
        r"""Effect-sizes parameter."""
        return self.get('effsizes')

    @effsizes.setter
    def effsizes(self, v):
        self.set('effsizes', ascontiguousarray(v))



"""
********************
Covariance functions
********************

- :class:`.LinearCov`
- :class:`.SumCov`
- :class:`.EyeCov`

Linear covariance
^^^^^^^^^^^^^^^^^

.. autoclass:: LinearCov
  :members:

Sum covariance
^^^^^^^^^^^^^^

.. autoclass:: SumCov
  :members:

Identity covariance
^^^^^^^^^^^^^^^^^^^

.. autoclass:: EyeCov
  :members:
"""

from ._linear import LinearCov
from ._sum import SumCov
from ._eye import EyeCov
from ._free import FreeFormCov
